@@ -46,8 +46,8 @@ else:
     EAGAIN = errno.EAGAIN
 
 VERSION_MAJOR=0
-VERSION_MINOR=4
-VERSION_REVISION=93
+VERSION_MINOR=9
+VERSION_REVISION=0
 VERSION_NUMBER=(VERSION_MAJOR*1000000+VERSION_MINOR*1000+VERSION_REVISION)
 
 MQTTv31 = 3
@@ -197,11 +197,22 @@ def topic_matches_sub(sub, topic):
     slen = len(sub)
     tlen = len(topic)
 
+    if slen > 0 and tlen > 0:
+        if (sub[0] == '$' and topic[0] != '$') or (topic[0] == '$' and sub[0] != '$'):
+            return False
+
     spos = 0
     tpos = 0
 
     while spos < slen and tpos < tlen:
         if sub[spos] == topic[tpos]:
+            if tpos == tlen-1:
+                # Check for e.g. foo matching foo/#
+                if spos == slen-3 and sub[spos+1] == '/' and sub[spos+2] == '#':
+                    result = True
+                    multilevel_wildcard = True
+                    break
+
             spos += 1
             tpos += 1
 
@@ -231,13 +242,6 @@ def topic_matches_sub(sub, topic):
                 result = False
                 break
 
-        if tpos == tlen-1:
-            # Check for e.g. foo matching foo/#
-            if spos == slen-3 and sub[spos+1] == '/' and sub[spos+2] == '#':
-                result = True
-                multilevel_wildcard = True
-                break
-
     if not multilevel_wildcard and (tpos < tlen or spos < slen):
         result = False
 
@@ -257,8 +261,7 @@ def _socketpair_compat():
     try:
         sock1.connect(("localhost", port))
     except socket.error as err:
-        (msg) = err
-        if msg.errno != errno.EINPROGRESS:
+        if err.errno != errno.EINPROGRESS and err.errno != errno.EWOULDBLOCK and err.errno != EAGAIN:
             raise
     sock2, address = listensock.accept()
     sock2.setblocking(0)
@@ -438,6 +441,7 @@ class Client(object):
         self.on_connect = None
         self.on_publish = None
         self.on_message = None
+        self.on_message_filtered = []
         self.on_subscribe = None
         self.on_unsubscribe = None
         self.on_log = None
@@ -461,6 +465,7 @@ class Client(object):
         self._tls_ca_certs = None
         self._tls_cert_reqs = None
         self._tls_ciphers = None
+        self._tls_version = tls_version
         self._tls_insecure = False
 
     def __del__(self):
@@ -709,8 +714,7 @@ class Client(object):
             else:
                 self._sock = socket.create_connection((self._host, self._port), source_address=(self._bind_address, 0))
         except socket.error as err:
-            (msg) = err
-            if msg.errno != errno.EINPROGRESS:
+            if err.errno != errno.EINPROGRESS and err.errno != errno.EWOULDBLOCK and err.errno != EAGAIN:
                 raise
 
         if self._tls_ca_certs is not None:
@@ -790,7 +794,7 @@ class Client(object):
             try:
                 self._sockpairR.recv(1)
             except socket.error as err:
-                if err.errno != errno.EAGAIN:
+                if err.errno != EAGAIN:
                     raise
 
         if self.socket() in socklist[1]:
@@ -1255,6 +1259,45 @@ class Client(object):
         self._thread.join()
         self._thread = None
 
+    def message_callback_add(self, sub, callback):
+        """Register a message callback for a specific topic.
+        Messages that match 'sub' will be passed to 'callback'. Any
+        non-matching messages will be passed to the default on_message
+        callback.
+        
+        Call multiple times with different 'sub' to define multiple topic
+        specific callbacks.
+        
+        Topic specific callbacks may be removed with
+        message_callback_remove()."""
+        if callback is None or sub is None:
+            raise ValueError("sub and callback must both be defined.")
+
+        self._callback_mutex.acquire()
+
+        for i in range(0, len(self.on_message_filtered)):
+            if self.on_message_filtered[i][0] == sub:
+                self.on_message_filtered[i] = (sub, callback)
+                self._callback_mutex.release()
+                return
+
+        self.on_message_filtered.append((sub, callback))
+        self._callback_mutex.release()
+
+    def message_callback_remove(self, sub):
+        """Remove a message callback previously registered with
+        message_callback_add()."""
+        if sub is None:
+            raise ValueError("sub must defined.")
+
+        self._callback_mutex.acquire()
+        for i in range(0, len(self.on_message_filtered)):
+            if self.on_message_filtered[i][0] == sub:
+                self.on_message_filtered.pop(i)
+                self._callback_mutex.release()
+                return
+        self._callback_mutex.release()
+
     # ============================================================
     # Private functions
     # ============================================================
@@ -1302,12 +1345,11 @@ class Client(object):
                 else:
                     command = self._sock.recv(1)
             except socket.error as err:
-                (msg) = err
-                if self._ssl and (msg.errno == ssl.SSL_ERROR_WANT_READ or msg.errno == ssl.SSL_ERROR_WANT_WRITE):
+                if self._ssl and (err.errno == ssl.SSL_ERROR_WANT_READ or err.errno == ssl.SSL_ERROR_WANT_WRITE):
                     return MQTT_ERR_AGAIN
-                if msg.errno == EAGAIN:
+                if err.errno == EAGAIN:
                     return MQTT_ERR_AGAIN
-                print(msg)
+                print(err)
                 return 1
             else:
                 if len(command) == 0:
@@ -1326,12 +1368,11 @@ class Client(object):
                     else:
                         byte = self._sock.recv(1)
                 except socket.error as err:
-                    (msg) = err
-                    if self._ssl and (msg.errno == ssl.SSL_ERROR_WANT_READ or msg.errno == ssl.SSL_ERROR_WANT_WRITE):
+                    if self._ssl and (err.errno == ssl.SSL_ERROR_WANT_READ or err.errno == ssl.SSL_ERROR_WANT_WRITE):
                         return MQTT_ERR_AGAIN
-                    if msg.errno == EAGAIN:
+                    if err.errno == EAGAIN:
                         return MQTT_ERR_AGAIN
-                    print(msg)
+                    print(err)
                     return 1
                 else:
                     byte = struct.unpack("!B", byte)
@@ -1358,12 +1399,11 @@ class Client(object):
                 else:
                     data = self._sock.recv(self._in_packet['to_process'])
             except socket.error as err:
-                (msg) = err
-                if self._ssl and (msg.errno == ssl.SSL_ERROR_WANT_READ or msg.errno == ssl.SSL_ERROR_WANT_WRITE):
+                if self._ssl and (err.errno == ssl.SSL_ERROR_WANT_READ or err.errno == ssl.SSL_ERROR_WANT_WRITE):
                     return MQTT_ERR_AGAIN
-                if msg.errno == EAGAIN:
+                if err.errno == EAGAIN:
                     return MQTT_ERR_AGAIN
-                print(msg)
+                print(err)
                 return 1
             else:
                 self._in_packet['to_process'] = self._in_packet['to_process'] - len(data)
@@ -1405,12 +1445,11 @@ class Client(object):
                 return MQTT_ERR_SUCCESS
             except socket.error as err:
                 self._current_out_packet_mutex.release()
-                (msg) = err
-                if self._ssl and (msg.errno == ssl.SSL_ERROR_WANT_READ or msg.errno == ssl.SSL_ERROR_WANT_WRITE):
+                if self._ssl and (err.errno == ssl.SSL_ERROR_WANT_READ or err.errno == ssl.SSL_ERROR_WANT_WRITE):
                     return MQTT_ERR_AGAIN
-                if msg.errno == EAGAIN:
+                if err.errno == EAGAIN:
                     return MQTT_ERR_AGAIN
-                print(msg)
+                print(err)
                 return 1
 
             if write_length > 0:
@@ -1804,7 +1843,7 @@ class Client(object):
         try:
             self._sockpairW.send(sockpair_data)
         except socket.error as err:
-            if err.errno != errno.EAGAIN:
+            if err.errno != EAGAIN:
                 raise
 
         if not self._in_callback and self._thread is None:
@@ -1932,23 +1971,11 @@ class Client(object):
 
         message.timestamp = time.time()
         if message.qos == 0:
-            self._callback_mutex.acquire()
-            if self.on_message:
-                self._in_callback = True
-                self.on_message(self, self._userdata, message)
-                self._in_callback = False
-
-            self._callback_mutex.release()
+            self._handle_on_message(message)
             return MQTT_ERR_SUCCESS
         elif message.qos == 1:
             rc = self._send_puback(message.mid)
-            self._callback_mutex.acquire()
-            if self.on_message:
-                self._in_callback = True
-                self.on_message(self, self._userdata, message)
-                self._in_callback = False
-
-            self._callback_mutex.release()
+            self._handle_on_message(message)
             return rc
         elif message.qos == 2:
             rc = self._send_pubrec(message.mid)
@@ -1978,12 +2005,7 @@ class Client(object):
 
                 # Only pass the message on if we have removed it from the queue - this
                 # prevents multiple callbacks for the same message.
-                self._callback_mutex.acquire()
-                if self.on_message:
-                    self._in_callback = True
-                    self.on_message(self, self._userdata, self._in_messages[i])
-                    self._in_callback = False
-                self._callback_mutex.release()
+                self._handle_on_message(self._in_messages[i])
                 self._in_messages.pop(i)
                 self._inflight_messages = self._inflight_messages - 1
                 if self._max_inflight_messages > 0:
@@ -2091,6 +2113,23 @@ class Client(object):
         self._out_message_mutex.release()
         return MQTT_ERR_SUCCESS
 
+    def _handle_on_message(self, message):
+        self._callback_mutex.acquire()
+        matched = False
+        for t in self.on_message_filtered:
+            if topic_matches_sub(t[0], message.topic):
+                self._in_callback = True
+                t[1](self, self._userdata, message)
+                self._in_callback = False
+                matched = True
+
+        if matched == False and self.on_message:
+            self._in_callback = True
+            self.on_message(self, self._userdata, message)
+            self._in_callback = False
+
+        self._callback_mutex.release()
+
     def _thread_main(self):
         run = True
         self._thread_terminate = False
